/************************************************************************
 *                                                                      *
 *    Copyright (c) 1991 Wilhelm Valder, Regine Freitag, Wolfgang Dzida *
 *                        All rights reserved                           *
 *                                                                      *
 *                                                                      *
 *    Permission to copy this program or any part of this program is    *
 *    granted in case of personal, non-commercial use only. Any use     *
 *    for profit or other commercial gain without written permissions   *
 *    of the authors is prohibited.                                     *
 *                                                                      *
 *    You may freely distribute this program to friends, colleagues,    *
 *    etc.  As a courtesy we would appreciate it if you left our names  *
 *    on the code as the authors.                                       *
 *    We also, of course, disavow any liabilities                       *
 *    for any bizarre things you may choose to do with the system.      *
 *                                                                      *
 *    We would appreciate though your letting us know about any         *
 *    extensions, modifications, project suggestions, etc.              *
 *    Please mail all bugs, comments, complaints and suggestions to:    *
 *                                                                      *
 *                                                                      *
 *    GMD (German National Research Center for Computer Science)        *
 *    Institut fuer Systemtechnik (F2G2)                                *
 *    P.O. Box 1240                                                     *
 *    Schloss Birlinghoven                                              *
 *    D-5205 St. Augustin 1, West Germany                               *
 *                                                                      *
 *    email:    dzida@kmx.gmd.dbp.de                                    *
 *              freitag@gmd.de                                          *
 *                                                                      *
 ************************************************************************/
/****************************************************************************
 * This file contains the functions to access the fonts and pixmaps of the
 * ERGO-shell
 ****************************************************************************/

#include <stdio.h>
#include <X11/StringDefs.h>
#include <Xm/Xm.h>
#include "exLayout.h"
#include <exFonts.h>
#include <exErrors.h>
#include <pixels/shadow.bits>

static char SccsID[] = "@(#)exLayout.c	1.3\t1/18/93";            /* sccs keyword */

static Pixmap _exWhitePM, _exBlackPM, _exGreyPM, _exLight_greyPM, 
       _exDark_greyPM, _exUltralight_greyPM; 
                            /* internal variables for pixmaps of ERGO-shell */
static Pixmap _exLight_greyBM;
                            /* internal variables for bitmaps of ERGO-shell */
static XFontStruct *_exSmallFont, *_exBaseFont, *_exBoldFont, *_exEditorFont;
                            /* internal variables for fonts of ERGO-shell */

static exInitFonts();
static exInitPixmaps();
static exInitBitmaps();

/***************************************************************************
 * public routines: 
 **************************************************************************/

/* exInitLayout ************************************************************
 * initializes font resources and pixmaps
 **************************************************************************/

int exInitLayout(display)
     Display *display;
{
  exInitFonts(display);
  return(exInitPixmaps(display));
}

/* exReset **************************************************************
 * closes resource fonts and frees pixmaps
 **************************************************************************/

void exReset(display)
     Display *display;
{  
  /* close fonts */
  if (_exBaseFont)
    XFreeFont(display, _exBaseFont);
  if (_exBoldFont)
    XFreeFont(display, _exBoldFont);
  if (_exEditorFont)
    XFreeFont(display, _exEditorFont);
  if (_exSmallFont)
    XFreeFont(display, _exSmallFont);

  /* free pixmaps */
  if (_exBlackPM)
    XFreePixmap(display, _exBlackPM);
  if (_exWhitePM)
    XFreePixmap(display, _exWhitePM);
  if (_exGreyPM)
    XFreePixmap(display, _exGreyPM);
  if(_exLight_greyPM)
    XFreePixmap(display, _exLight_greyPM);
  if (_exDark_greyPM)
    XFreePixmap(display, _exDark_greyPM);
  if (_exUltralight_greyPM)
    XFreePixmap(display, _exUltralight_greyPM);

  /* free bitmaps */
  if (_exLight_greyBM)
    XFreePixmap(display, _exLight_greyBM);
}

/* getMouse_xy: **************************************************************
 * provides coordinates of mouse position in root window
 ****************************************************************************/

getMouse_xy(d, rootx, rooty)
     Display *d;
     int *rootx, *rooty;        /* coordinates relative to root window */
{
  Window rootw, subw;
  int x, y;                     /* coordinates relative to specified window */
  unsigned int  mask;

  if (! XQueryPointer(d, DefaultRootWindow(d), &rootw, &subw, rootx, rooty, 
		      &x, &y, &mask)) {
    *rootx = 0;
    *rooty = 0;
    return(setError(ERR_NULL));
  }
  
  return(setError(ERR_NOERR));
}

/****************************************************************************
 *                           font handling 
 ****************************************************************************/

/* exFontHeight ************************************************************
 * calculates bounding box height of a font. (The bounding box height of a 
 * font is determined by adding the portion of the font below the baseline 
 * to the portion at and above the baseline.)
 **************************************************************************/

exFontHeight(fontInfo)
     XFontStruct *fontInfo;  /* (in) pointer to font information */
{
 return(fontInfo->max_bounds.ascent + fontInfo->max_bounds.descent); 
}

/* exFontWidth ************************************************************
 * calculates bounding box width of a font
 **************************************************************************/

exFontWidth(fontInfo)
     XFontStruct *fontInfo;
{
  return(fontInfo->max_bounds.rbearing + fontInfo->min_bounds.lbearing);
}

/* exBaseFont, exBoldFont, exEditorFont, exSmallFont: **********************
 * provides font information of baseFont, boldFont, editorFont, smallFont
 **************************************************************************/

XFontStruct *exBaseFont(display)
     Display *display;
{
  if (! _exBaseFont)
    exInitFonts(display);
  return(_exBaseFont);
}

XFontStruct *exBoldFont(display)
     Display *display;
{
  if (! _exBoldFont)
    exInitFonts(display);
  return(_exBoldFont);
}

XFontStruct *exEditorFont(display)
     Display *display;
{
  if (! _exEditorFont)
    exInitFonts(display);
  return(_exEditorFont);
}

XFontStruct *exSmallFont(display)
     Display *display;
{
  if (! _exSmallFont)
    exInitFonts(display);
  return(_exSmallFont);
}

/* getFont: ******************************************************************
 * retrieves the font associated with a fontlist
 ****************************************************************************/

XFontStruct *getFont(fl)
     XmFontList fl;
{
  XmFontContext context;
  XmStringCharSet charset;
  XFontStruct *font;

  if (! XmFontListInitFontContext(&context, fl))
    return(NULL);

  XmFontListGetNextFont(context, &charset, &font);

  XmFontListFreeFontContext(context);

  return(font);
}

/****************************************************************************
 * pixmaps (on color screens: depth > 1)
 ****************************************************************************/

/* exWhitePM, exBlackPM, exGreyPM, exLight_greyPM, exDark_greyPM, exUltralight_greyPM: *************************************
 * provides white, black, grey, light grey, dark grey, ultralight grey pixmap
 ****************************************************************************/

Pixmap exWhitePM(display)
     Display *display;
{
  if (_exWhitePM == (Pixmap) 0)
    exInitPixmaps(display);
  return(_exWhitePM);
}

Pixmap exBlackPM(display)
     Display *display;
{
  if ( _exBlackPM == (Pixmap) 0)
    exInitPixmaps(display);
  return(_exBlackPM);
}

Pixmap exGreyPM(display)
     Display *display;
{
  if ( _exGreyPM == (Pixmap) 0)
    exInitPixmaps(display);
  return(_exGreyPM);
}

Pixmap exLight_greyPM(display)
     Display *display;
{
  if ( _exLight_greyPM == (Pixmap) 0)
    exInitPixmaps(display);
  return(_exLight_greyPM);
}

Pixmap exDark_greyPM(display)
     Display *display;
{
  if (_exDark_greyPM == (Pixmap) 0)
    exInitPixmaps(display);
  return(_exDark_greyPM);
}

Pixmap exUltralight_greyPM(display)
     Display *display;
{
  if (_exUltralight_greyPM == (Pixmap) 0)
    exInitPixmaps(display);
  return(_exUltralight_greyPM);
}

/* exWhitePixel, exBlackPixel: *********************************************
 * return the corresponding pixel value
 **************************************************************************/

unsigned long exWhitePixel(display)
     Display *display;
{
  return(XWhitePixel(display, XDefaultScreen(display)));
}

unsigned long exBlackPixel(display)
     Display *display;
{
  return(XBlackPixel(display, XDefaultScreen(display)));
}

/* exLoadPixmap: *************************************************************
 * loads a bitmap from the given file in ESHBASEDIR; then determines the
 * optimal size and generates a pixmap from this bitmap
 ****************************************************************************/

exLoadPixmap(d, file, pm)
     Display *d;
     char *file;
     Pixmap *pm;
{
  unsigned int w, h, rw, rh;
  int xh, yh, depth;
  Pixmap data, pix_data;
  char fullName[256];
  Window win;
  GC util_gc;
  char *pixeldir;
  char *getenv();

  if (file[0] != '/') {                           /* file in pixeldir */
    if (!(pixeldir = getenv("ESHBASEDIR"))) 
      fatal("exLoadPixmap", 
	    "Environment variable ESHBASEDIR not defined: exiting!", NULL);

    makeFullName(pixeldir, "include/pixels/", fullName, 256);
    strcat(fullName, file);
  }
  else                                            /* absolute pathname */
    strcpy(fullName, file);

  win = DefaultRootWindow(d);
  if (XReadBitmapFile(d, (Drawable) win, (char *) fullName, &w, &h, 
		      &data, &xh, &yh) != BitmapSuccess) {
    *pm = (Pixmap) NULL;
    exError("exLoadPixmap", "cannot load bitmap from file -- ", 
	    					fullName, " --", NULL);
    return(setError(ERR_READ));
  }

  XQueryBestTile(d, (Drawable) win, w,h, &rw, &rh); 
  if (rw > w || rh > h) {         
    /* if efficient sizes are bigger than pixmap size, display of the pixmap 
       isn't supported best */
    *pm = (Pixmap) NULL;
    return(setError(ERR_ALLOC));
  } 
  
  depth = DefaultDepth(d, DefaultScreen(d));
  pix_data = XCreatePixmap(d, (Drawable) win, w, h, depth);
  if (pix_data == None) {
    *pm = (Pixmap) NULL;
    exError("exLoadPixmap", "cannot create pixmap from file -- ", 
	    					fullName, " --", NULL);
    return(setError(ERR_READ));
  }

  util_gc = XCreateGC(d, pix_data, 0, 0);
  XSetBackground(d, util_gc, exWhitePixel(d));
  XSetForeground(d, util_gc, exBlackPixel(d));
  XSetGraphicsExposures(d, util_gc, False);
  XCopyPlane(d, data, pix_data, util_gc, 0, 0, w, h, 0, 0, 1);
  XFreePixmap(d, data);
  XFreeGC(d, util_gc);

  *pm = pix_data;
  if (*pm == (Pixmap) NULL)
    return(setError(ERR_ABORT));

  BYE;
}

/* exDrawShadow: ************************************************************
 * This procedure is used to draw boxes with shadows underneath them, like 
 * the box around a pull-down menu or a notifier.
 ****************************************************************************/

void exDrawShadow(wd, x, y, width, height)
    Widget wd;			/* Widget in which to draw shadowed box. */
    int x, y;			/* UL corner of shadowed box (gives location
				 * of outside pixel of box/shadow). */
    unsigned int width, height;	/* Dimensions of useable area INSIDE
				 * shadowed box. */
{
  static Pixmap shadowPixmap = 0;
  static GC fill_gc = NULL, draw_gc = NULL;
  Display *d = XtDisplay(wd);
  Window w = XtWindow(wd);
  XRectangle rects[2];

  if (! fill_gc) {
    fill_gc = XCreateGC(d, w, 0, 0);
    XSetForeground(d, fill_gc, exBlackPixel(d));
    XSetBackground(d, fill_gc, exWhitePixel(d));
    XSetFillStyle(d, fill_gc, FillStippled);
  }

  if (shadowPixmap == 0) {
    unsigned int pm_width, pm_height;
    XQueryBestTile(d, w, shadow_width, shadow_height, &pm_width, &pm_height);
    
    if ((pm_width <= 16) && (pm_height <= 16))
      shadowPixmap =  XCreateBitmapFromData(d, w, 
			(char *) shadow_bits, pm_width, pm_height);
      
    if (shadowPixmap == 0)
      shadowPixmap = exBlackPM(d);
    XSetStipple(d, fill_gc, shadowPixmap);
  }
  
  if (! draw_gc) {
    draw_gc = XCreateGC(d, w, 0, 0);
    XSetForeground(d, draw_gc, exBlackPixel(d));
    XSetBackground(d, draw_gc, exWhitePixel(d));
    XSetFillStyle(d, draw_gc, FillSolid);
  }
    
/*
  XDrawRectangle(d, w, draw_gc, x, y, width, height);
*/
  rects[0].x = x + width;
  rects[0].y = y + SHADOW_TOP;
  rects[0].width = SHADOW_WIDTH;
  rects[0].height = height + SHADOW_WIDTH - SHADOW_TOP;
  
  rects[1].x = x + SHADOW_WIDTH;
  rects[1].y = y + height;
  rects[1].width = width - SHADOW_TOP;
  rects[1].height = SHADOW_WIDTH;
  
  XFillRectangles(d, w, fill_gc, rects, 2);
}

/****************************************************************************
 * bitmaps (always depth = 1)
 ****************************************************************************/

/* exLight_greyBM: **********************************************************
 * provides light grey bitmap
 ****************************************************************************/

Pixmap exLight_greyBM(display)
     Display *display;
{
  if (_exLight_greyBM == (Pixmap) 0)
    exInitBitmaps(display);
  return(_exLight_greyBM);
}

/* exLoadBitmap: *************************************************************
 * loads a bitmap from the given file in ESHBASEDIR
 ****************************************************************************/

exLoadBitmap(d, file, pm)
     Display *d;
     char *file;
     Pixmap *pm;
{
  unsigned int w, h, rw, rh;
  int xh, yh;
  Pixmap data;
  char fullName[256];
  Window win;
  char *pixeldir;
  char *getenv();

  if (file[0] != '/') {                           /* file in pixeldir */
    if (!(pixeldir = getenv("ESHBASEDIR"))) 
      fatal("exLoadPixmap", 
	    "Environment variable ESHBASEDIR not defined: exiting!", NULL);

    makeFullName(pixeldir, "include/pixels/", fullName, 256);
    strcat(fullName, file);
  }
  else                                            /* absolute pathname */
    strcpy(fullName, file);

  win = DefaultRootWindow(d);
  if (XReadBitmapFile(d, (Drawable) win, (char *) fullName, &w, &h, 
		      &data, &xh, &yh) != BitmapSuccess) {
    *pm = (Pixmap) NULL;
    exError("exLoadPixmap", "cannot load bitmap from file -- ", 
	    					fullName, " --", NULL);
    return(setError(ERR_READ));
  }

  XQueryBestTile(d, (Drawable) win, w,h, &rw, &rh); 
  if (rw > w || rh > h) {         
    /* if efficient sizes are bigger than pixmap size, display of the pixmap 
       isn't supported best */
    *pm = (Pixmap) NULL;
    return(setError(ERR_ALLOC));
  } 
  
  *pm = data;
  if (*pm == (Pixmap) NULL)
    return(setError(ERR_ABORT));

  BYE;
}

/****************************************************************************
 * private routines 
 ****************************************************************************/

/* exInitFonts ************************************************************
 * initialize fonts
 **************************************************************************/

static exInitFonts(display)
     Display *display;      /* (in) */
{
  if (!(_exBaseFont = XLoadQueryFont(display, EXBASEFONT)))
    return(setError(ERR_INIT));

  if (!(_exEditorFont = XLoadQueryFont(display, EXEDITORFONT)))
    return(setError(ERR_INIT));

  if (!(_exBoldFont = XLoadQueryFont(display, EXBOLDFONT)))
    return(setError(ERR_INIT));

  if (!(_exSmallFont = XLoadQueryFont(display, EXSMALLFONT)))
    return(setError(ERR_INIT));

  return(setError(ERR_NOERR));
}

/* exInitPixmaps: ***********************************************************
 * loads pixmaps from files if not available; 
 * returns error code if reading fails
 ****************************************************************************/

static exInitPixmaps(display)
     Display *display;
{
  register rc = 0;
  
  if ((rc = exLoadPixmap(display, "black.bmp", &_exBlackPM)) 
      != ERR_NOERR)
    return(rc);
  if ((rc = exLoadPixmap(display, "white.bmp", &_exWhitePM)) 
      != ERR_NOERR)
    return(rc);
  if ((_exGreyPM = XmGetPixmap(XDefaultScreenOfDisplay(display), 
			       "50_foreground", 1, 0)) 
      == (Pixmap) 0)
    return(setError(ERR_ALLOC));
  if ((_exLight_greyPM = XmGetPixmap(XDefaultScreenOfDisplay(display), 
				     "25_foreground", 1, 0))
      == (Pixmap) 0)
    return(setError(ERR_ALLOC));
  if ((_exDark_greyPM = XmGetPixmap(XDefaultScreenOfDisplay(display), 
				    "75_foreground", 1, 0))
      == (Pixmap) 0)
    return(setError(ERR_ALLOC));
  if ((rc = exLoadPixmap(display, "ultralight-grey.bmp",&_exUltralight_greyPM))
      != ERR_NOERR)
    return(rc);
  
  return(ERR_NOERR);
}

/* exInitBitmaps: ***********************************************************
 * loads pixmaps from files if not available; 
 * returns error code if reading fails
 ****************************************************************************/

static exInitBitmaps(display)
     Display *display;
{
  register rc = 0;
  
  if ((rc = exLoadBitmap(display, "light-grey.bmp", &_exLight_greyBM)) 
      != ERR_NOERR)
    return(rc);
  
  return(ERR_NOERR);
}

